#ifndef _DEMON_H
#define _DEMON_H


extern "C" {
#include <signal.h>   // signal
#include <stdio.h>    // perror
#include <errno.h>    // perror
#include <string.h>  // strncmp

#ifdef WIN32
#pragma warning(disable: 4786)
#include <io.h>
#include <time.h>
#include <sys/timeb.h>
#else
  // UNIX
#include <unistd.h>   // select
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/time.h>  // time
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#include <stdlib.h>   // atoi
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
}


#include "general.h"

#include <set>
#include <tcl.h>
#include <iostream> // cout
#include <fstream>

#include "common.h"
#include "Member.h"
#include "SMPHeader.h"
#include "Operation.h"
#include "BlockedADU.h"
#include "Timer.h"
#include "MemberSet.h"
#include "DADU.h"
#include "timerlist.h"
#include "ratectrlqueue.h"
#include "grouplist.h"
#include "dadulist.h"

#ifdef _SMP_SENDDELAYED
#define DELAY_TIME 150
#endif


//
// smoothing factor for delay estimates
//
#define ALPHA 0.6


namespace SMP {

const char version[] = "2.1.5";

/*********************************************************************
 Name:      Client
 Desc:      keeps all information related to the clients of smp
 *********************************************************************/
class Client {
public:
  Client();
  ~Client();

  int actNumber;                             /* current no. of clients */
  bool belegt[MAX_CLIENTS];                  /* occupied client slots */
  bool loggedOn[MAX_CLIENTS];                /* client joined a group? */
  bool accepted[MAX_CLIENTS];                /* client accepted by smp? */
  int msock_r[MAX_CLIENTS];                  /* mc rec. sockets of clients */	
  int msock_w[MAX_CLIENTS];                  /* mc write sockets of clients */
  struct sockaddr_in mc_addr_r[MAX_CLIENTS]; /* socket addr for mc receive */
  struct sockaddr_in mc_addr_w[MAX_CLIENTS]; /* socket addr for mc write */
  struct ip_mreq mreq[MAX_CLIENTS];          /* needed to join mc group */
  //SORTLIST<Member> grouplist[MAX_CLIENTS]; 
  GroupList grouplist[MAX_CLIENTS];          /* keeps track of group members */
  std::set<uint> CollisionList[MAX_CLIENTS];     /* keeps ssrcs of collisions */
  unsigned int seqNumber[MAX_CLIENTS];       /* stores current seq.no. */
  Member* We[MAX_CLIENTS];                   /* points to ourselfes */
  int Members[MAX_CLIENTS];                  /* stores group size */
  double AvgSMSize[MAX_CLIENTS];             /* tracks avg. size of sess.msg.*/
  double ServiceClassFactor[MAX_CLIENTS];    /* stores service class */
  double SessionBandwidth[MAX_CLIENTS];      /* session bandwidth */

  /* smp parameter for new clients */
  bool SourceOrdering[MAX_CLIENTS];          /* source ordering ? */
  bool LateJoin[MAX_CLIENTS];                /* late join ? */
  bool SharedData[MAX_CLIENTS];              /* shared data ? */
  bool Multicast[MAX_CLIENTS];               /* multicast/unicast ? */
  //bool Loopback[MAX_CLIENTS];                /* multicast loopback ? */

  /* delayed adu mechansim (application lj) */
  bool OpenRequest[MAX_CLIENTS];             /* open dadu request (lj) */
  bool FirstLateJoinInd[MAX_CLIENTS];        /* pass only first msg. to app. */
  struct DelayedADURequestBuffer_t {         /* store data of  dadu request */
    bool BufferFree;
    unsigned int ReceiverSSRC;               /* sender ssrc */
    char * data;                             /* shm position */
    int size;
  } DADURequest_ReceiveBuffer[MAX_CLIENTS],  /* received dadu requests */
    DADURequest_SendBuffer[MAX_CLIENTS];     /* send dadu requests     */


  unsigned int dadu_seqno[MAX_CLIENTS];      /* id counter for dadus */
  DADU DAduRequest[MAX_CLIENTS];             /* DADU request of client, needed
					        for REPEAT timer action */
  DAduList dadulist[MAX_CLIENTS];

  //SORTLIST<DADU> dadulist[MAX_CLIENTS];  
  /* list of received delayed adu 
     requests, needed for 
     GIVETHROUGH timer action */
  
  DADU * FindDADU(int, unsigned int);        /* find DADU with specific id */

  /* rate control */
  double Credit[MAX_CLIENTS];                  /* current client credit */
  int MaxKbps[MAX_CLIENTS];                    /* max. data rate */
  RateCtrlQueue ratecontrolqueue[MAX_CLIENTS]; /* rate control list */
  //SORTLIST<BlockedADU> ratecontrolqueue[MAX_CLIENTS];

  /* Local recovery Varicales */
  unsigned char GlobalTTL[MAX_CLIENTS];

  MemberSet* MSet[MAX_CLIENTS];              /* set of active members */

  /* Check Liveness of Members */
  int SMCount[MAX_CLIENTS];
  unsigned int TimeOfLastCheck[MAX_CLIENTS];

  /* look up toHandle*/
  int getNetsockPosition(int fd);
  
  /* get client if only the socket fd is known*/
  int  search_pos();
};

/*********************************************************************
 Name:      Demon
 Desc:      main class, manages network io and client io
 *********************************************************************/
class Demon {
 public:
  enum { 
    FIRSTTIME = 0,
    DEFAULTDISTANCE = 50,     // default distance to other members in ms 
    MAX_RETRIES = 100         // max. times that is tried to request a lost packet (no source-ordering only)
  };
  
  static Demon* Deamon;

  static void SignalHandler(...);   /* timer handler */
  static void ShutDown(int);  /* smoothly terminates smp when we receive SIGTERM, etc. */

  Demon();

  void logOn(int, LogOnInfo*);  /* initiated by SMP::JoinGroup */
  void logOff(int);             /* initiated by SMP::LeaveGroup */
  void close(int);              /* close connection to client */

  int OpenReceiveSocket(int, LogOnInfo*);  /* open receive socket */
  int OpenSendSocket(int, LogOnInfo*);     /* open send socket */
  int Reconnect(int);  /* reconnect socket ... when using unicast with no recipant */

  // receive data
  static void HandleIncomingData(ClientData, int);    /* incoming net traffic */

  int ReceiveData(int, char*, int, sockaddr_in);
  int ReceiveUnreliableData(int, char*, int, sockaddr_in);
  int ReceiveJoin(int, char*, int, sockaddr_in);
  int ReceiveLeave(int, char*, int, sockaddr_in);
  int ReceiveRequest(int, char*, int, sockaddr_in);
  int ReceiveSessionMsg(int, char*, int, sockaddr_in);
  int ReceiveDADUReq(int, char*, int, sockaddr_in);
  int ReceiveDADURsp(int, char*, int, sockaddr_in);

  void IndData(int, char*, int, unsigned int);  /* indicate to client that data has been received */
  void IndUnreliableData(int, char*, int, unsigned int);  /* indicate to client that unreliable data has been received */
  void IndClose(int, unsigned int);  /* indicates to the client that we left the group */
  void IndLateJoin(int, unsigned int);  /* indicates to the client that a late join would be neccessary */
  void IndDADUReq(int, char*, int, unsigned int, unsigned int);  /* indicates to the client that a delayed adu req has been received */
  void IndDADURsp(int, char*, int, unsigned int);  /* indicates to the client that a delayed adu rsp has been received */
  void IndDADUReqFailed(int);  /* indicates to the client that the delayed adu request was not answered */

  // send data
  int SendJoinMsg(int);
  int SendLeaveMsg(int);
  int SendData(int, void*, int, bool = true);
  int SendUnreliableData(int, void*, int, bool = true);
  int SendRepair(int, void*, int, unsigned char);
  int SendRequest(int, unsigned int, unsigned int);
  int _SendSessionMsg(int, unsigned int, unsigned int);
  int SendSessionMsg(int, unsigned int, unsigned int);
  int SendDADUReq(int, void*, int, bool = true);
  int SendDADURsp(int, void*, int, unsigned int, bool = true);

  void RateControl(int);  /* check rate control list and send packets */

  bool sendCtlMsg(int, void*, int, bool = true);  /* send ctrl. message (Join, Leave, Request, Sessionmessage) */
  int AddSessionMessageBody(char*, int);  /* build session message */
  void ComputeSessionMessage(int, char*, unsigned int, unsigned int,  Member*);  /* ckeck incoming session message */

  // timer functions
  int SetAlarmTimer();  /* initialize the alarm timer */
  Timer* FindTimer(int, int, unsigned int, unsigned int);
  Timer* DeleteTimer(int, int, unsigned int, unsigned int);
  void DeleteAllTimerForLeavingApplication(int, unsigned int);
  void DeleteRunningTimer();  /* clear system's interval timer */
  void ExecuteTimerAction(Timer*);
  void SetTimer(int, unsigned int, unsigned int, int, int, Member*,
                unsigned int, double, bool, unsigned char, bool, int, Timer* = NULL);

  // helper functions
  unsigned int GetDistanceInMillis(unsigned int, unsigned int);  /* estimate distance in ms (according to simplified NTP) */
  unsigned int GetTimestamp();  /* gets time in ms */

  unsigned int GetSSRCIdentifier();  /* creates a new ssrc */
  void SSRCCollision(int, unsigned int);  /* ssrc collision treatment according to RFC 1889 */
  void UpdateTimerList(unsigned int, unsigned int);  /* update timers after ssrc collision */

  unsigned int GetSequenceNumber(int);  /* set seq.no. of a data packet */

  void UpdateAckList(int, unsigned int, unsigned int, bool = true);  /* check ack list if no late join desired */

  Client c;                          /* client related info */
  int sock;                  /* socket for accepting clients */
#ifdef WIN32
  struct sockaddr_in addr;  
#else
  struct sockaddr_un addr;           /* socket addr. */
#endif
  int HostIpAddr;                    /* ip addr in network representation */
  Tcl_TimerToken timer;
  static Timer* Current;             /* current timer (first due) */
  bool mc_loop;                      /* multicast loop back option */
  static TimerList timerlist;        /* list of all timers set */
  //static SORTLIST<Timer> timerlist;

  /* some masks for signal & io handling */

  //unix
  //static sigset_t mask;

  void CheckLivenessOfMembers(int);  /* check members periodically */
  const int CHECKPERIOD;  // CHECKPERIOD == 60 => each 5 min.
                          // CHECKPERIOD == 120 => each 10 min.


  /* remember last call to signal handler function */
  unsigned int last_call;
  Member* CheckMember(int, unsigned int, unsigned int, int, unsigned int,
		      int, bool, bool*);

  char* AllocatePacketBuffer(int, int);
  void FreePacketBuffer(int, char*, int, bool = true);

  Operation* InsertOpList(int, Member*, void*, int, unsigned int, bool = true);
  bool DeleteOpList(int, Member*, unsigned int, bool = true);
  void UpdateOpList(int, Member*, unsigned int, unsigned int, bool = true);
  void UpdateAllOpLists(int, unsigned int, unsigned int, bool = true);

  char recv_buf[RECV_BUF];
  char send_buf[10000];
};
}


#endif







